@@ -1,6 +1,5 @@
 /*
- * A mount(8) for Linux 0.99.
- * mount.c,v 1.1.1.1 1993/11/18 08:40:51 jrs Exp
+ * A mount(8) for Linux.
  *
  * Modifications by many people. Distributed under GPL.
  */
@@ -22,14 +21,15 @@
 #include <sys/mount.h>
 
 #include <mntent.h>
-#include <paths.h>
 
 #ifdef HAVE_LIBSELINUX
 #include <selinux/selinux.h>
 #include <selinux/context.h>
 #endif
 
+#include "pathnames.h"
 #include "fsprobe.h"
+#include "devname.h"
 #include "mount_constants.h"
 #include "sundries.h"
 #include "xmalloc.h"
@@ -37,11 +37,10 @@
 #include "fstab.h"
 #include "lomount.h"
 #include "loop.h"
-#include "linux_fs.h"		/* for BLKGETSIZE */
 #include "getusername.h"
-#include "mount_paths.h"
 #include "env.h"
 #include "nls.h"
+#include "blkdev.h"
 
 #define DO_PS_FIDDLING
 
@@ -49,32 +48,26 @@
 #include "setproctitle.h"
 #endif
 
-/* Quiet mode */
-int mount_quiet = 0;
-
 /* True for fake mount (-f).  */
 static int fake = 0;
 
 /* True if we are allowed to call /sbin/mount.${FSTYPE} */
 static int external_allowed = 1;
 
-/* Don't write a entry in /etc/mtab (-n).  */
+/* Don't write an entry in /etc/mtab (-n).  */
 static int nomtab = 0;
 
 /* True for explicit readonly (-r).  */
 static int readonly = 0;
 
-/* Nonzero for chatty (-v).  */
-int verbose = 0;
-
 /* Nonzero for sloppy (-s).  */
-int sloppy = 0;
+static int sloppy = 0;
 
 /* True for explicit read/write (-w).  */
 static int readwrite = 0;
 
 /* True for all mount (-a).  */
-int mount_all = 0;
+static int mount_all = 0;
 
 /* True for fork() during all mount (-F).  */
 static int optfork = 0;
@@ -82,17 +75,26 @@ static int optfork = 0;
 /* Add volumelabel in a listing of mounted devices (-l). */
 static int list_with_volumelabel = 0;
 
-/* Nonzero for mount {--bind|--replace|--before|--after|--over|--move|
- * 		       make-shared|make-private|make-unbindable|make-slave}
+/* Nonzero for mount {bind|move|make-shared|make-private|
+ *				make-unbindable|make-slave}
  */
 static int mounttype = 0;
 
-/* True if ruid != euid.  */
-static int suid = 0;
+/* True if (ruid != euid) or (0 != ruid), i.e. only "user" mounts permitted.  */
+static int restricted = 1;
 
 /* Contains the fd to read the passphrase from, if any. */
 static int pfd = -1;
 
+/* mount(2) options */
+struct mountargs {
+       const char *spec;
+       const char *node;
+       const char *type;
+       int flags;
+       void *data;
+};
+
 /* Map from -o and fstab option strings to the flag argument to mount(2).  */
 struct opt_map {
   const char *opt;		/* option name */
@@ -172,6 +174,10 @@ static const struct opt_map opt_map[] = {
   { "atime",	0, 1, MS_NOATIME },	/* Update access time */
   { "noatime",	0, 0, MS_NOATIME },	/* Do not update access time */
 #endif
+#ifdef MS_I_VERSION
+  { "iversion",	0, 0, MS_I_VERSION },	/* Update inode I_version time */
+  { "noiversion", 0, 1, MS_I_VERSION },	/* Don't update inode I_version time */
+#endif
 #ifdef MS_NODIRATIME
   { "diratime",	0, 1, MS_NODIRATIME },	/* Update dir access times */
   { "nodiratime", 0, 0, MS_NODIRATIME },/* Do not update dir access times */
@@ -182,11 +188,18 @@ static const struct opt_map opt_map[] = {
   { "norelatime", 0, 1, MS_RELATIME }, /* Update access time without regard
 					  to mtime/ctime */
 #endif
+#ifdef MS_STRICTATIME
+  { "strictatime", 0, 0, MS_STRICTATIME }, /* Strict atime semantics */
+  { "nostrictatime", 0, 1, MS_STRICTATIME }, /* kernel default atime */
+#endif
+  { "nofail",	0, 0, MS_COMMENT},	/* Do not fail if ENOENT on dev */
   { NULL,	0, 0, 0		}
 };
 
-static const char *opt_loopdev, *opt_vfstype, *opt_offset, *opt_encryption,
-	*opt_speed, *opt_comment, *opt_uhelper;
+static int opt_nofail = 0;
+
+static const char *opt_loopdev, *opt_vfstype, *opt_offset, *opt_sizelimit,
+        *opt_encryption, *opt_speed, *opt_comment, *opt_uhelper;
 
 static int mounted (const char *spec0, const char *node0);
 static int check_special_mountprog(const char *spec, const char *node,
@@ -200,6 +213,7 @@ static struct string_opt_map {
   { "loop=",	0, &opt_loopdev },
   { "vfs=",	1, &opt_vfstype },
   { "offset=",	0, &opt_offset },
+  { "sizelimit=",  0, &opt_sizelimit },
   { "encryption=", 0, &opt_encryption },
   { "speed=", 0, &opt_speed },
   { "comment=", 1, &opt_comment },
@@ -230,12 +244,6 @@ parse_string_opt(char *s) {
 	return 0;
 }
 
-static void
-my_free(const void *s) {
-	if (s)
-		free((void *) s);
-}
-
 /* Report on a single mount.  */
 static void
 print_one (const struct my_mntent *me) {
@@ -246,8 +254,8 @@ print_one (const struct my_mntent *me) {
 		printf (" type %s", me->mnt_type);
 	if (me->mnt_opts != NULL)
 		printf (" (%s)", me->mnt_opts);
-	if (list_with_volumelabel) {
-		const char *devname = fsprobe_get_devname(me->mnt_fsname);
+	if (list_with_volumelabel && is_pseudo_fs(me->mnt_type) == 0) {
+		const char *devname = spec_to_devname(me->mnt_fsname);
 
 		if (devname) {
 			const char *label;
@@ -335,7 +343,7 @@ append_context(const char *optname, char *optdata, char **extra_opts)
 	security_context_t raw = NULL;
 	char *data = NULL;
 
-	if (!is_selinux_enabled())
+	if (is_selinux_enabled() != 1)
 		/* ignore the option if we running without selinux */
 		return 0;
 
@@ -346,8 +354,8 @@ append_context(const char *optname, char *optdata, char **extra_opts)
 	data = *optdata =='"' ? strip_quotes(optdata) : optdata;
 
 	if (selinux_trans_to_raw_context(
-			(security_context_t) data, &raw)==-1 ||
-			raw==NULL)
+			(security_context_t) data, &raw) == -1 ||
+			raw == NULL)
 		return -1;
 
 	if (verbose)
@@ -389,6 +397,8 @@ parse_opt(char *opt, int *mask, char **extra_opts) {
 				verbose = 0;
 			}
 #endif
+			if (streq(opt, "nofail"))
+				opt_nofail = 1;
 			return;
 		}
 
@@ -425,6 +435,10 @@ parse_opt(char *opt, int *mask, char **extra_opts) {
 		if (append_context("defcontext=", opt+11, extra_opts) == 0)
 			return;
 	}
+	if (strncmp(opt, "rootcontext=", 12) == 0 && *(opt+12)) {
+		if (append_context("rootcontext=", opt+12, extra_opts) == 0)
+			return;
+	}
 #endif
 	*extra_opts = append_opt(*extra_opts, opt, NULL);
 }
@@ -502,9 +516,11 @@ fix_opts_string (int flags, const char *extra_opts, const char *user) {
 }
 
 static int
-already (const char *spec, const char *node) {
+already (const char *spec0, const char *node0) {
 	struct mntentchn *mc;
 	int ret = 1;
+	char *spec = canonicalize_spec(spec0);
+	char *node = canonicalize(node0);
 
 	if ((mc = getmntfile(node)) != NULL)
 		error (_("mount: according to mtab, "
@@ -516,6 +532,10 @@ already (const char *spec, const char *node) {
 		       spec, mc->m.mnt_dir);
 	else
 		ret = 0;
+
+	free(spec);
+	free(node);
+
 	return ret;
 }
 
@@ -529,11 +549,11 @@ create_mtab (void) {
 
 	lock_mtab();
 
-	mfp = my_setmntent (MOUNTED, "a+");
+	mfp = my_setmntent (_PATH_MOUNTED, "a+");
 	if (mfp == NULL || mfp->mntent_fp == NULL) {
 		int errsv = errno;
 		die (EX_FILEIO, _("mount: can't open %s for writing: %s"),
-		     MOUNTED, strerror (errsv));
+		     _PATH_MOUNTED, strerror (errsv));
 	}
 
 	/* Find the root entry by looking it up in fstab */
@@ -541,16 +561,16 @@ create_mtab (void) {
 		char *extra_opts;
 		parse_opts (fstab->m.mnt_opts, &flags, &extra_opts);
 		mnt.mnt_dir = "/";
-		mnt.mnt_fsname = fsprobe_get_devname(fstab->m.mnt_fsname);
+		mnt.mnt_fsname = spec_to_devname(fstab->m.mnt_fsname);
 		mnt.mnt_type = fstab->m.mnt_type;
 		mnt.mnt_opts = fix_opts_string (flags, extra_opts, NULL);
 		mnt.mnt_freq = mnt.mnt_passno = 0;
-		my_free(extra_opts);
+		free(extra_opts);
 
 		if (my_addmntent (mfp, &mnt) == 1) {
 			int errsv = errno;
 			die (EX_FILEIO, _("mount: error writing %s: %s"),
-			     MOUNTED, strerror (errsv));
+			     _PATH_MOUNTED, strerror (errsv));
 		}
 	}
 	if (fchmod (fileno (mfp->mntent_fp), 0644) < 0)
@@ -558,11 +578,13 @@ create_mtab (void) {
 			int errsv = errno;
 			die (EX_FILEIO,
 			     _("mount: error changing mode of %s: %s"),
-			     MOUNTED, strerror (errsv));
+			     _PATH_MOUNTED, strerror (errsv));
 		}
 	my_endmntent (mfp);
 
 	unlock_mtab();
+
+	reset_mtab_info();
 }
 
 /* count successful mount system calls */
@@ -624,7 +646,10 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
   if (!external_allowed)
       return 0;
 
-  if (type && strlen(type) < 100) {
+  if (type == NULL || strcmp(type, "none") == 0)
+	  return 0;
+
+  if (strlen(type) < 100) {
        sprintf(mountprog, "/sbin/mount.%s", type);
        if (stat(mountprog, &statbuf) == 0) {
 	    if (verbose)
@@ -660,7 +685,7 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
 
 		 if (verbose > 2) {
 			i = 0;
-			while(verbose > 2 && mountargs[i]) {
+			while(mountargs[i]) {
 				printf("mount: external mount: argv[%d] = \"%s\"\n",
 					i, mountargs[i]);
 				i++;
@@ -685,6 +710,157 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
 }
 
 
+/* list of already tested filesystems by procfsloop_mount() */
+static struct tried {
+	struct tried *next;
+	char *type;
+} *tried = NULL;
+
+static int
+was_tested(const char *fstype) {
+	struct tried *t;
+
+	if (fsprobe_known_fstype(fstype))
+		return 1;
+	for (t = tried; t; t = t->next) {
+		if (!strcmp(t->type, fstype))
+			return 1;
+	}
+	return 0;
+}
+
+static void
+set_tested(const char *fstype) {
+	struct tried *t = xmalloc(sizeof(struct tried));
+
+	t->next = tried;
+	t->type = xstrdup(fstype);
+	tried = t;
+}
+
+static void
+free_tested(void) {
+	struct tried *t, *tt;
+
+	t = tried;
+	while(t) {
+		free(t->type);
+		tt = t->next;
+		free(t);
+		t = tt;
+	}
+	tried = NULL;
+}
+
+static char *
+procfsnext(FILE *procfs) {
+   char line[100];
+   char fsname[100];
+
+   while (fgets(line, sizeof(line), procfs)) {
+      if (sscanf (line, "nodev %[^\n]\n", fsname) == 1) continue;
+      if (sscanf (line, " %[^ \n]\n", fsname) != 1) continue;
+      return xstrdup(fsname);
+   }
+   return 0;
+}
+
+/* Only use /proc/filesystems here, this is meant to test what
+   the kernel knows about, so /etc/filesystems is irrelevant.
+   Return: 1: yes, 0: no, -1: cannot open procfs */
+static int
+known_fstype_in_procfs(const char *type)
+{
+    FILE *procfs;
+    char *fsname;
+    int ret = -1;
+
+    procfs = fopen(_PATH_PROC_FILESYSTEMS, "r");
+    if (procfs) {
+	ret = 0;
+	while ((fsname = procfsnext(procfs)) != NULL)
+	    if (!strcmp(fsname, type)) {
+		ret = 1;
+		break;
+	    }
+	fclose(procfs);
+	procfs = NULL;
+    }
+    return ret;
+}
+
+/* Try all types in FILESYSTEMS, except those in *types,
+   in case *types starts with "no" */
+/* return: 0: OK, -1: error in errno, 1: type not found */
+/* when 0 or -1 is returned, *types contains the type used */
+/* when 1 is returned, *types is NULL */
+static int
+procfsloop_mount(int (*mount_fn)(struct mountargs *, int *, int *),
+			 struct mountargs *args,
+			 const char **types,
+			 int *special, int *status)
+{
+	char *files[2] = { _PATH_FILESYSTEMS, _PATH_PROC_FILESYSTEMS };
+	FILE *procfs;
+	char *fsname;
+	const char *notypes = NULL;
+	int no = 0;
+	int ret = 1;
+	int errsv = 0;
+	int i;
+
+	if (*types && !strncmp(*types, "no", 2)) {
+		no = 1;
+		notypes = (*types) + 2;
+	}
+	*types = NULL;
+
+	/* Use _PATH_PROC_FILESYSTEMS only when _PATH_FILESYSTEMS
+	 * (/etc/filesystems) does not exist.  In some cases trying a
+	 * filesystem that the kernel knows about on the wrong data will crash
+	 * the kernel; in such cases _PATH_FILESYSTEMS can be used to list the
+	 * filesystems that we are allowed to try, and in the order they should
+	 * be tried.  End _PATH_FILESYSTEMS with a line containing a single '*'
+	 * only, if _PATH_PROC_FILESYSTEMS should be tried afterwards.
+	 */
+	for (i=0; i<2; i++) {
+		procfs = fopen(files[i], "r");
+		if (!procfs)
+			continue;
+		while ((fsname = procfsnext(procfs)) != NULL) {
+			if (!strcmp(fsname, "*")) {
+				fclose(procfs);
+				goto nexti;
+			}
+			if (was_tested (fsname))
+				continue;
+			if (no && matching_type(fsname, notypes))
+				continue;
+			set_tested (fsname);
+			args->type = fsname;
+			if (verbose)
+				printf(_("Trying %s\n"), fsname);
+			if ((*mount_fn) (args, special, status) == 0) {
+				*types = fsname;
+				ret = 0;
+				break;
+			} else if (errno != EINVAL &&
+				   known_fstype_in_procfs(fsname) == 1) {
+				*types = "guess";
+				ret = -1;
+				errsv = errno;
+				break;
+			}
+		}
+		free_tested();
+		fclose(procfs);
+		errno = errsv;
+		return ret;
+	nexti:;
+	}
+	return 1;
+}
+
 static const char *
 guess_fstype_by_devname(const char *devname)
 {
@@ -695,8 +871,8 @@ guess_fstype_by_devname(const char *devname)
 
       if (!type)
          printf (_("       I will try all types mentioned in %s or %s\n"),
-		      ETC_FILESYSTEMS, PROC_FILESYSTEMS);
-      else if (!strcmp(type, "swap"))
+		      _PATH_FILESYSTEMS, _PATH_PROC_FILESYSTEMS);
+      else if (!strcmp(type, MNTTYPE_SWAP))
          printf (_("       and it looks like this is swapspace\n"));
       else
          printf (_("       I will try type %s\n"), type);
@@ -719,13 +895,10 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
    if (*types && strcasecmp (*types, "auto") == 0)
       *types = NULL;
 
-   if (!*types && (flags & (MS_BIND | MS_MOVE)))
-      *types = "none";		/* random, but not "bind" */
-
    if (!*types && !(flags & MS_REMOUNT)) {
       *types = guess_fstype_by_devname(spec);
       if (*types) {
-	  if (!strcmp(*types, "swap")) {
+	  if (!strcmp(*types, MNTTYPE_SWAP)) {
 	      error(_("%s looks like swapspace - not mounted"), spec);
 	      *types = NULL;
 	      return 1;
@@ -758,16 +931,16 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
       return do_mount (&args, special, status);
    }
 
-   return fsprobe_procfsloop_mount(do_mount, &args, types, special, status);
+   return procfsloop_mount(do_mount, &args, types, special, status);
 }
 
 /*
- * suid_check()
+ * restricted_check()
  *	Die if the user is not allowed to do this.
  */
 static void
-suid_check(const char *spec, const char *node, int *flags, char **user) {
-  if (suid) {
+restricted_check(const char *spec, const char *node, int *flags, char **user) {
+  if (restricted) {
       /*
        * MS_OWNER: Allow owners to mount when fstab contains
        * the owner option.  Note that this should never be used
@@ -828,11 +1001,53 @@ suid_check(const char *spec, const char *node, int *flags, char **user) {
   *flags &= ~(MS_OWNER | MS_GROUP);
 }
 
+/* Check, if there already exists a mounted loop device on the mountpoint node
+ * with the same parameters.
+ */
+static int
+is_mounted_same_loopfile(const char *node0, const char *loopfile, unsigned long long offset)
+{
+	struct mntentchn *mnt = NULL;
+	char *node;
+	int res = 0;
+
+	node = canonicalize(node0);
+
+	/* Search for mountpoint node in mtab,
+	 * procceed if any of these has the loop option set or
+	 * the device is a loop device
+	 */
+	mnt = getmntdirbackward(node, mnt);
+	if (!mnt) {
+		free(node);
+		return 0;
+	}
+	for(; mnt && res == 0; mnt = getmntdirbackward(node, mnt)) {
+		char *p;
+
+		if (strncmp(mnt->m.mnt_fsname, "/dev/loop", 9) == 0)
+			res = loopfile_used_with((char *) mnt->m.mnt_fsname,
+					loopfile, offset);
+
+		else if ((p = strstr(mnt->m.mnt_opts, "loop="))) {
+			char *dev = xstrdup(p+5);
+			if ((p = strchr(dev, ',')))
+				*p = '\0';
+			res = loopfile_used_with(dev, loopfile, offset);
+			free(dev);
+		}
+	}
+
+	free(node);
+	return res;
+}
+
 static int
 loop_check(const char **spec, const char **type, int *flags,
-	   int *loop, const char **loopdev, const char **loopfile) {
+	   int *loop, const char **loopdev, const char **loopfile,
+	   const char *node) {
   int looptype;
-  unsigned long long offset;
+  unsigned long long offset, sizelimit;
 
   /*
    * In the case of a loop mount, either type is of the form lo@/dev/loop5
@@ -857,7 +1072,7 @@ loop_check(const char **spec, const char **type, int *flags,
       *type = opt_vfstype;
   }
 
-  *loop = ((*flags & MS_LOOP) || *loopdev || opt_offset || opt_encryption);
+  *loop = ((*flags & MS_LOOP) || *loopdev || opt_offset || opt_sizelimit || opt_encryption);
   *loopfile = *spec;
 
   if (*loop) {
@@ -866,10 +1081,19 @@ loop_check(const char **spec, const char **type, int *flags,
       if (verbose)
 	printf(_("mount: skipping the setup of a loop device\n"));
     } else {
-      int loopro = (*flags & MS_RDONLY);
+      int loop_opts = SETLOOP_AUTOCLEAR; /* always attempt autoclear */
       int res;
 
+      if (*flags & MS_RDONLY)
+        loop_opts |= SETLOOP_RDONLY;
+
       offset = opt_offset ? strtoull(opt_offset, NULL, 0) : 0;
+      sizelimit = opt_sizelimit ? strtoull(opt_sizelimit, NULL, 0) : 0;
+
+      if (is_mounted_same_loopfile(node, *loopfile, offset)) {
+        error(_("mount: according to mtab %s is already mounted on %s as loop"), *loopfile, node);
+        return EX_FAIL;
+      }
 
       do {
         if (!*loopdev || !**loopdev)
@@ -879,8 +1103,8 @@ loop_check(const char **spec, const char **type, int *flags,
 	if (verbose)
 	  printf(_("mount: going to use the loop device %s\n"), *loopdev);
 
-	if ((res = set_loop(*loopdev, *loopfile, offset,
-			    opt_encryption, pfd, &loopro))) {
+	if ((res = set_loop(*loopdev, *loopfile, offset, sizelimit,
+			    opt_encryption, pfd, &loop_opts))) {
 	  if (res == 2) {
 	     /* loop dev has been grabbed by some other process,
 	        try again, if not given explicitly */
@@ -909,8 +1133,13 @@ loop_check(const char **spec, const char **type, int *flags,
       if (verbose > 1)
 	printf(_("mount: setup loop device successfully\n"));
       *spec = *loopdev;
-      if (loopro)
-	*flags |= MS_RDONLY;
+
+      if (loop_opts & SETLOOP_RDONLY)
+        *flags |= MS_RDONLY;
+
+      if (loop_opts & SETLOOP_AUTOCLEAR)
+        /* Prevent recording loop dev in mtab for cleanup on umount */
+        *loop = 0;
     }
   }
 
@@ -922,7 +1151,7 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 		  const char *opts, int flags, int freq, int pass) {
 	struct my_mntent mnt;
 
-	mnt.mnt_fsname = canonicalize (spec);
+	mnt.mnt_fsname = is_pseudo_fs(type) ? xstrdup(spec) : canonicalize(spec);
 	mnt.mnt_dir = canonicalize (node);
 	mnt.mnt_type = type;
 	mnt.mnt_opts = opts;
@@ -934,6 +1163,13 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 	if (verbose)
 		print_one (&mnt);
 
+	if (!nomtab && mtab_does_not_exist()) {
+		if (verbose > 1)
+			printf(_("mount: no %s found - creating it..\n"),
+			       _PATH_MOUNTED);
+		create_mtab ();
+	}
+
 	if (!nomtab && mtab_is_writable()) {
 		if (flags & MS_REMOUNT)
 			update_mtab (mnt.mnt_dir, &mnt);
@@ -943,16 +1179,16 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 			mntFILE *mfp;
 
 			lock_mtab();
-			mfp = my_setmntent(MOUNTED, "a+");
+			mfp = my_setmntent(_PATH_MOUNTED, "a+");
 			if (mfp == NULL || mfp->mntent_fp == NULL) {
 				int errsv = errno;
-				error(_("mount: can't open %s: %s"), MOUNTED,
+				error(_("mount: can't open %s: %s"), _PATH_MOUNTED,
 				      strerror (errsv));
 			} else {
 				if ((my_addmntent (mfp, &mnt)) == 1) {
 					int errsv = errno;
 					error(_("mount: error writing %s: %s"),
-					      MOUNTED, strerror (errsv));
+					      _PATH_MOUNTED, strerror (errsv));
 				}
 			}
 			my_endmntent(mfp);
@@ -991,16 +1227,14 @@ cdrom_setspeed(const char *spec) {
 
 /*
  * try_mount_one()
- *	Try to mount one file system. When "bg" is 1, this is a retry
- *	in the background. One additional exit code EX_BG is used here.
- *	It is used to instruct the caller to retry the mount in the
- *	background.
+ *	Try to mount one file system.
+ *
  * returns: 0: OK, EX_SYSERR, EX_FAIL, return code from nfsmount,
  *      return status from wait
  */
 static int
 try_mount_one (const char *spec0, const char *node0, const char *types0,
-	       const char *opts0, int freq, int pass, int bg, int ro) {
+	       const char *opts0, int freq, int pass, int ro) {
   int res = 0, status = 0, special = 0;
   int mnt5_res = 0;		/* only for gcc */
   int mnt_err;
@@ -1012,6 +1246,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   int loop = 0;
   const char *loopdev = 0, *loopfile = 0;
   struct stat statbuf;
+  int retries = 0;	/* Nr of retries for mount in case of ENOMEDIUM */
 
   /* copies for freeing on exit */
   const char *opts1, *spec1, *node1, *types1, *extra_opts1;
@@ -1035,7 +1270,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   if (mount_all && (flags & MS_NOAUTO))
       goto out;
 
-  suid_check(spec, node, &flags, &user);
+  restricted_check(spec, node, &flags, &user);
 
   /* The "mount -f" checks for for existing record in /etc/mtab (with
    * regular non-fake mount this is usually done by kernel)
@@ -1056,7 +1291,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
        * stale assignments of files to loop devices. Nasty when used for
        * encryption.
        */
-      res = loop_check(&spec, &types, &flags, &loop, &loopdev, &loopfile);
+      res = loop_check(&spec, &types, &flags, &loop, &loopdev, &loopfile, node);
       if (res)
 	  goto out;
   }
@@ -1064,6 +1299,9 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   if (loop)
       opt_loopdev = loopdev;
 
+  if (flags & (MS_BIND | MS_MOVE | MS_PROPAGATION))
+      types = "none";
+
   /*
    * Call mount.TYPE for types that require a separate mount program.
    * For the moment these types are ncpfs and smbfs. Maybe also vxfs.
@@ -1074,6 +1312,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       goto out;
   }
 
+mount_retry:
   block_signals (SIG_BLOCK);
 
   if (!fake) {
@@ -1115,7 +1354,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   /* Mount failed, complain, but don't die.  */
 
   if (types == 0) {
-    if (suid)
+    if (restricted)
       error (_("mount: I could not determine the filesystem type, "
 	       "and none was specified"));
     else
@@ -1154,9 +1393,11 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       else if (stat (node, &statbuf))
 	   error (_("mount: mount point %s is a symbolic link to nowhere"),
 		  node);
-      else if (stat (spec, &statbuf))
+      else if (stat (spec, &statbuf)) {
+	   if (opt_nofail)
+		goto out;
 	   error (_("mount: special device %s does not exist"), spec);
-      else {
+      } else {
 	   errno = mnt_err;
 	   perror("mount");
       }
@@ -1164,53 +1405,49 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
     case ENOTDIR:
       if (stat (node, &statbuf) || ! S_ISDIR(statbuf.st_mode))
 	   error (_("mount: mount point %s is not a directory"), node);
-      else if (stat (spec, &statbuf) && errno == ENOTDIR)
+      else if (stat (spec, &statbuf) && errno == ENOTDIR) {
+	   if (opt_nofail)
+              goto out;
 	   error (_("mount: special device %s does not exist\n"
 		    "       (a path prefix is not a directory)\n"), spec);
-      else {
+      } else {
 	   errno = mnt_err;
 	   perror("mount");
       }
       break;
     case EINVAL:
     { int fd;
-      unsigned long size;
-      int warned=0;
+      unsigned long long size = 0;
 
       if (flags & MS_REMOUNT) {
 	error (_("mount: %s not mounted already, or bad option"), node);
       } else {
 	error (_("mount: wrong fs type, bad option, bad superblock on %s,\n"
-	       "       missing codepage or other error"),
+	       "       missing codepage or helper program, or other error"),
 	       spec);
 
-	if (stat(spec, &statbuf) == 0 && S_ISBLK(statbuf.st_mode)
-	   && (fd = open(spec, O_RDONLY | O_NONBLOCK)) >= 0) {
-	  if (ioctl(fd, BLKGETSIZE, &size) == 0) {
-	    if (size == 0 && !loop) {
-	      warned++;
+	if (stat(spec, &statbuf) < 0) {
+	  if (errno == ENOENT)         /* network FS? */
+	    error(_(
+	       "       (for several filesystems (e.g. nfs, cifs) you might\n"
+	       "       need a /sbin/mount.<type> helper program)"));
+
+	} else if (S_ISBLK(statbuf.st_mode)
+	                 && (fd = open(spec, O_RDONLY | O_NONBLOCK)) >= 0) {
+
+	  if (blkdev_get_size(fd, &size) == 0) {
+	    if (size == 0 && !loop)
 	      error(_(
 		 "       (could this be the IDE device where you in fact use\n"
 		 "       ide-scsi so that sr0 or sda or so is needed?)"));
-	    }
-	    if (size && size <= 2) {
-	      warned++;
+
+	    if (size && size <= 2)
 	      error(_(
 		  "       (aren't you trying to mount an extended partition,\n"
 		  "       instead of some logical partition inside?)"));
-	    }
-	  close(fd);
+
+	    close(fd);
 	  }
-#if 0
-	  /* 0xf for SCSI, 0x3f for IDE. One might check /proc/partitions
-	     to see whether this thing really is partitioned.
-	     Do not suggest partitions for /dev/fd0. */
-	  if (!warned && (statbuf.st_rdev & 0xf) == 0) {
-	    warned++;
-	    error ("       (could this be the whole disk device\n"
-		   "       where you need a partition?)");
-	  }
-#endif
 	}
 	error(_(
 		"       In some cases useful info is found in syslog - try\n"
@@ -1224,7 +1461,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       error (_("mount: %s: can't read superblock"), spec); break;
     case ENODEV:
     {
-      int pfs = fsprobe_known_fstype_in_procfs(types);
+      int pfs = known_fstype_in_procfs(types);
 
       if (pfs == 1 || !strcmp(types, "guess"))
         error(_("mount: %s: unknown device"), spec);
@@ -1243,13 +1480,13 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	    u++;
 	  }
 	}
-	if (u && fsprobe_known_fstype_in_procfs(lowtype) == 1)
+	if (u && known_fstype_in_procfs(lowtype) == 1)
 	  error (_("mount: probably you meant %s"), lowtype);
 	else if (!strncmp(lowtype, "iso", 3) &&
-			fsprobe_known_fstype_in_procfs("iso9660") == 1)
+			known_fstype_in_procfs("iso9660") == 1)
 	  error (_("mount: maybe you meant 'iso9660'?"));
 	else if (!strncmp(lowtype, "fat", 3) &&
-			fsprobe_known_fstype_in_procfs("vfat") == 1)
+			known_fstype_in_procfs("vfat") == 1)
 	  error (_("mount: maybe you meant 'vfat'?"));
 	free(lowtype);
       } else
@@ -1258,6 +1495,8 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       break;
     }
     case ENOTBLK:
+      if (opt_nofail)
+        goto out;
       if (stat (spec, &statbuf)) /* strange ... */
 	error (_("mount: %s is not a block device, and stat fails?"), spec);
       else if (S_ISBLK(statbuf.st_mode))
@@ -1270,6 +1509,8 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	error (_("mount: %s is not a block device"), spec);
       break;
     case ENXIO:
+      if (opt_nofail)
+        goto out;
       error (_("mount: %s is not a valid block device"), spec); break;
     case EACCES:  /* pre-linux 1.1.38, 1.1.41 and later */
     case EROFS:   /* linux 1.1.38 and later */
@@ -1282,6 +1523,10 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	  error (_("mount: %s%s is write-protected but explicit `-w' flag given"),
 		 bd, spec);
 	  break;
+      } else if (flags & MS_REMOUNT) {
+	  error (_("mount: cannot remount %s%s read-write, is write-protected"),
+		 bd, spec);
+	  break;
       } else {
 	 opts = opts0;
 	 types = types0;
@@ -1296,11 +1541,22 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	     types = 0;
          error (_("mount: %s%s is write-protected, mounting read-only"),
 		bd, spec0);
-	 res = try_mount_one (spec0, node0, types, opts, freq, pass, bg, 1);
+	 res = try_mount_one (spec0, node0, types, opts, freq, pass, 1);
 	 goto out;
       }
       break;
     }
+    case ENOMEDIUM:
+      if (retries < CRDOM_NOMEDIUM_RETRIES) {
+	      if (verbose)
+		      printf(_("mount: no medium found on %s ...trying again\n"),
+				 spec);
+              sleep(3);
+	      ++retries;
+              goto mount_retry;
+      }
+      error(_("mount: no medium found on %s"), spec);
+      break;
     default:
       error ("mount: %s", strerror (mnt_err)); break;
     }
@@ -1308,6 +1564,27 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   res = EX_FAIL;
 
  out:
+
+#if defined(HAVE_LIBSELINUX) && defined(HAVE_SECURITY_GET_INITIAL_CONTEXT)
+  if (res != EX_FAIL && verbose && is_selinux_enabled() > 0) {
+      security_context_t raw = NULL, def = NULL;
+
+      if (getfilecon(node, &raw) > 0 &&
+		     security_get_initial_context("file", &def) == 0) {
+
+	  if (!selinux_file_context_cmp(raw, def))
+	      printf(_("mount: %s does not contain SELinux labels.\n"
+                   "       You just mounted an file system that supports labels which does not\n"
+                   "       contain labels, onto an SELinux box. It is likely that confined\n"
+                   "       applications will generate AVC messages and not be allowed access to\n"
+                   "       this file system.  For more details see restorecon(8) and mount(8).\n"),
+                   node);
+      }
+      freecon(raw);
+      freecon(def);
+  }
+#endif
+
   my_free(extra_opts1);
   my_free(spec1);
   my_free(node1);
@@ -1317,19 +1594,6 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   return res;
 }
 
-/*
- * set_proc_name()
- *	Update the argument vector, so that this process may be easily
- *	identified in a "ps" listing.
- */
-static void
-set_proc_name (const char *spec)
-{
-#ifdef DO_PS_FIDDLING
-	setproctitle ("mount", spec);
-#endif
-}
-
 static char *
 subst_string(const char *s, const char *sub, int sublen, const char *repl) {
 	char *n;
@@ -1375,7 +1639,6 @@ is_existing_file (const char *s) {
 static int
 mount_one (const char *spec, const char *node, const char *types,
 	   const char *fstabopts, char *cmdlineopts, int freq, int pass) {
-	int status, status2;
 	const char *nspec;
 	char *opts;
 
@@ -1385,11 +1648,6 @@ mount_one (const char *spec, const char *node, const char *types,
 	/* Merge the fstab and command line options.  */
 	opts = append_opt(opts, cmdlineopts, NULL);
 
-	/* Handle possible LABEL= and UUID= forms of spec */
-	nspec = fsprobe_get_devname_for_mounting(spec);
-	if (nspec)
-		spec = nspec;
-
 	if (types == NULL && !mounttype && !is_existing_file(spec)) {
 		if (strchr (spec, ':') != NULL) {
 			types = "nfs";
@@ -1406,48 +1664,30 @@ mount_one (const char *spec, const char *node, const char *types,
 		}
 	}
 
-	/*
-	 * Try to mount the file system. When the exit status is EX_BG,
-	 * we will retry in the background. Otherwise, we're done.
-	 */
-	status = try_mount_one (spec, node, types, opts, freq, pass, 0, 0);
-	if (status != EX_BG)
-		return status;
+	/* Handle possible LABEL= and UUID= forms of spec */
+	if (types == NULL || (strncmp(types, "nfs", 3) &&
+			      strncmp(types, "cifs", 4) &&
+			      strncmp(types, "smbfs", 5))) {
+		nspec = spec_to_devname(spec);
+		if (nspec)
+			spec = nspec;
+	}
 
-	/*
-	 * Retry in the background.
-	 */
-	printf (_("mount: backgrounding \"%s\"\n"), spec);
-	fflush( stdout );		/* prevent duplicate output */
-	if (fork() > 0)
-		return 0;			/* parent returns "success" */
-	spec = xstrdup(spec);		/* arguments will be destroyed */
-	node = xstrdup(node);		/* by set_proc_name()          */
-	types = xstrdup(types);
-	set_proc_name (spec);		/* make a nice "ps" listing */
-	status2 = try_mount_one (spec, node, types, opts, freq, pass, 1, 0);
-	if (verbose && status2)
-		printf (_("mount: giving up \"%s\"\n"), spec);
-
-	my_free(opts);
-	my_free(node);
-	my_free(types);
-	exit (0);			/* child stops here */
+	return try_mount_one (spec, node, types, opts, freq, pass, 0);
 }
 
 /* Check if an fsname/dir pair was already in the old mtab.  */
 static int
 mounted (const char *spec0, const char *node0) {
 	struct mntentchn *mc, *mc0;
-	char *spec, *node;
+	const char *spec, *node;
 	int ret = 0;
 
 	/* Handle possible UUID= and LABEL= in spec */
-	spec0 = fsprobe_get_devname(spec0);
-	if (!spec0)
+	spec = spec_to_devname(spec0);
+	if (!spec)
 		return ret;
 
-	spec = canonicalize(spec0);
 	node = canonicalize(node0);
 
 	mc0 = mtab_head();
@@ -1626,14 +1866,10 @@ static struct option longopts[] = {
 	{ "test-opts", 1, 0, 'O' },
 	{ "pass-fd", 1, 0, 'p' },
 	{ "types", 1, 0, 't' },
-	{ "bind", 0, 0, 128 },
-	{ "replace", 0, 0, 129 },
-	{ "after", 0, 0, 130 },
-	{ "before", 0, 0, 131 },
-	{ "over", 0, 0, 132 },
-	{ "move", 0, 0, 133 },
+	{ "bind", 0, 0, 'B' },
+	{ "move", 0, 0, 'M' },
 	{ "guess-fstype", 1, 0, 134 },
-	{ "rbind", 0, 0, 135 },
+	{ "rbind", 0, 0, 'R' },
 	{ "make-shared", 0, 0, 136 },
 	{ "make-slave", 0, 0, 137 },
 	{ "make-private", 0, 0, 138 },
@@ -1684,10 +1920,7 @@ usage (FILE *fp, int n) {
 	  "Other options: [-nfFrsvw] [-o options] [-p passwdfd].\n"
 	  "For many more details, say  man 8 mount .\n"
 	));
-/*
-	  "Union or stack mounts are specified using one of\n"
-	  "       --replace, --after, --before, --over\n"
-*/
+
 	unlock_mtab();
 	exit (n);
 }
@@ -1699,6 +1932,9 @@ getfs(const char *spec, const char *uuid, const char *label)
 	struct mntentchn *mc = NULL;
 	const char *devname = NULL;
 
+	if (!spec && !uuid && !label)
+		return NULL;
+
 	/*
 	 * A) 99% of all cases, the spec on cmdline matches
 	 *    with spec in fstab
@@ -1724,7 +1960,7 @@ getfs(const char *spec, const char *uuid, const char *label)
 	else if (label)
 		devname = fsprobe_get_devname_by_label(label);
 	else
-		devname = fsprobe_get_devname(spec);
+		devname = spec_to_devname(spec);
 
 	if (devname)
 		mc = getfs_by_devname(devname);
@@ -1761,15 +1997,28 @@ getfs(const char *spec, const char *uuid, const char *label)
 	 *    Earlier mtab was tried first, but this would sometimes try the
 	 *    wrong mount in case mtab had the root device entry wrong.
 	 */
-	if (!mc)
+	if (!mc && (devname || spec))
 		mc = getmntfile (devname ? devname : spec);
 
-	if (devname)
-		my_free(devname);
+	my_free(devname);
 	return mc;
 }
 
-char *progname;
+
+static void
+print_version(int rc) {
+	printf(	"mount from %s (with "
+#ifdef HAVE_LIBBLKID
+		"libblkid"
+#else
+		"libvolume_id"
+#endif
+#ifdef HAVE_LIBSELINUX
+		" and selinux"
+#endif
+		" support)\n", PACKAGE_STRING);
+	exit(rc);
+}
 
 int
 main(int argc, char *argv[]) {
@@ -1807,12 +2056,15 @@ main(int argc, char *argv[]) {
 	initproctitle(argc, argv);
 #endif
 
-	while ((c = getopt_long (argc, argv, "afFhilL:no:O:p:rsU:vVwt:",
+	while ((c = getopt_long (argc, argv, "aBfFhilL:Mno:O:p:rRsU:vVwt:",
 				 longopts, NULL)) != -1) {
 		switch (c) {
 		case 'a':	       /* mount everything in fstab */
 			++mount_all;
 			break;
+		case 'B': /* bind */
+			mounttype = MS_BIND;
+			break;
 		case 'f':	       /* fake: don't actually call mount(2) */
 			++fake;
 			break;
@@ -1831,6 +2083,9 @@ main(int argc, char *argv[]) {
 		case 'L':
 			label = optarg;
 			break;
+		case 'M': /* move */
+			mounttype = MS_MOVE;
+			break;
 		case 'n':		/* do not write /etc/mtab */
 			++nomtab;
 			break;
@@ -1847,6 +2102,9 @@ main(int argc, char *argv[]) {
 			readonly = 1;
 			readwrite = 0;
 			break;
+		case 'R': /* rbind */
+			mounttype = (MS_BIND | MS_REC);
+			break;
 		case 's':		/* allow sloppy mount options */
 			sloppy = 1;
 			break;
@@ -1860,8 +2118,8 @@ main(int argc, char *argv[]) {
 			++verbose;
 			break;
 		case 'V':		/* version */
-			printf ("mount (%s)\n", PACKAGE_STRING);
-			exit (0);
+			print_version(EXIT_SUCCESS);
+			break;
 		case 'w':		/* mount read/write */
 			readwrite = 1;
 			readonly = 0;
@@ -1869,24 +2127,6 @@ main(int argc, char *argv[]) {
 		case 0:
 			break;
 
-		case 128: /* bind */
-			mounttype = MS_BIND;
-			break;
-		case 129: /* replace */
-			mounttype = MS_REPLACE;
-			break;
-		case 130: /* after */
-			mounttype = MS_AFTER;
-			break;
-		case 131: /* before */
-			mounttype = MS_BEFORE;
-			break;
-		case 132: /* over */
-			mounttype = MS_OVER;
-			break;
-		case 133: /* move */
-			mounttype = MS_MOVE;
-			break;
 		case 134:
 			/* undocumented, may go away again:
 			   call: mount --guess-fstype device
@@ -1898,9 +2138,6 @@ main(int argc, char *argv[]) {
 			printf("%s\n", fstype ? fstype : "unknown");
 			exit(fstype ? 0 : EX_FAIL);
 		    }
-		case 135:
-			mounttype = (MS_BIND | MS_REC);
-			break;
 
 		case 136:
 			mounttype = MS_SHARED;
@@ -1941,9 +2178,12 @@ main(int argc, char *argv[]) {
 	}
 
 	if (verbose > 2) {
-		printf("mount: fstab path: \"%s\"\n", _PATH_FSTAB);
-		printf("mount: lock path:  \"%s\"\n", MOUNTED_LOCK);
-		printf("mount: temp path:  \"%s\"\n", MOUNTED_TEMP);
+		printf("mount: fstab path: \"%s\"\n", _PATH_MNTTAB);
+		printf("mount: mtab path:  \"%s\"\n", _PATH_MOUNTED);
+		printf("mount: lock path:  \"%s\"\n", _PATH_MOUNTED_LOCK);
+		printf("mount: temp path:  \"%s\"\n", _PATH_MOUNTED_TMP);
+		printf("mount: UID:        %d\n", getuid());
+		printf("mount: eUID:       %d\n", geteuid());
 	}
 
 	argc -= optind;
@@ -1957,20 +2197,24 @@ main(int argc, char *argv[]) {
 		return print_all (types);
 	}
 
-	if (getuid () != geteuid ()) {
-		suid = 1;
-		if (types || options || readwrite || nomtab || mount_all ||
-		    fake || mounttype || (argc + specseen) != 1)
-			die (EX_USAGE, _("mount: only root can do that"));
+	{
+		const uid_t ruid = getuid();
+		const uid_t euid = geteuid();
+
+		/* if we're really root and aren't running setuid */
+		if (((uid_t)0 == ruid) && (ruid == euid)) {
+			restricted = 0;
+		}
 	}
 
-	if (!nomtab && mtab_does_not_exist()) {
-		if (verbose > 1)
-			printf(_("mount: no %s found - creating it..\n"),
-			       MOUNTED);
-		create_mtab ();
+	if (restricted &&
+	    (types || options || readwrite || nomtab || mount_all ||
+	     fake || mounttype || (argc + specseen) != 1)) {
+		die (EX_USAGE, _("mount: only root can do that"));
 	}
 
+	atexit(unlock_mtab);
+
 	switch (argc+specseen) {
 	case 0:
 		/* mount -a */
@@ -1981,19 +2225,24 @@ main(int argc, char *argv[]) {
 
 	case 1:
 		/* mount [-nfrvw] [-o options] special | node
+		 * mount -L label  (or -U uuid)
 		 * (/etc/fstab is necessary)
 		 */
 		if (types != NULL)
 			usage (stderr, EX_USAGE);
 
-		mc = getfs(*argv, uuid, label);
+		if (uuid || label)
+			mc = getfs(NULL, uuid, label);
+		else
+			mc = getfs(*argv, NULL, NULL);
+
 		if (!mc) {
 			if (uuid || label)
 				die (EX_USAGE, _("mount: no such partition found"));
 
 			die (EX_USAGE,
 			     _("mount: can't find %s in %s or %s"),
-			     *argv, _PATH_FSTAB, MOUNTED);
+			     *argv, _PATH_MNTTAB, _PATH_MOUNTED);
 		}
 
 		result = mount_one (xstrdup (mc->m.mnt_fsname),
